"""CLI utilities for MADSci.

This module provides common utilities for CLI commands.
"""

from madsci.client.cli.utils.output import error, get_console, info, success, warning

__all__ = [
    "error",
    "get_console",
    "info",
    "success",
    "warning",
]
